//
// plumbing - C++ wrapper facades for Berkeley sockets
// Copyright (C) 2009 Peter Miller
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License, version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <algorithm>

#include <libplumbing/at/service.h>
#include <libplumbing/at/job.h>


plumbing::at_service::~at_service()
{
    for (queue_t::iterator it = queue.begin(); it != queue.end(); ++it)
    {
        it->job->death_subscribe(0);
    }
}


plumbing::at_service::at_service()
{
}


void
plumbing::at_service::register_job_delta(const time_value &secs, at_job *job)
{
    register_job(time_value::now() + secs, job);
}


void
plumbing::at_service::register_job(const time_value &when, at_job *job)
{
    //
    // We want the job to unregister itself,
    // if its destructor is executed.
    //
    job->death_subscribe(this);

    queue_item item;
    item.when = when;
    item.job = job;

    //
    // We don't even check to see if it is there twice, because that is
    // OK (dumb, but OK).  Unless it kills itself.
    //
    for (queue_t::iterator it = queue.begin(); it != queue.end(); ++it)
    {
        if (item.when < it->when)
        {
            queue.insert(it, item);
            return;
        }
    }
    queue.push_back(item);
}


void
plumbing::at_service::unregister_job(at_job *job)
{
    //
    // Recall that a job can be present more than once.  We look for
    // more than one instance (especially of this is being called
    // from the job's destructor).
    //
    for (;;)
    {
        queue_t::iterator it = queue.begin();
        for (; it != queue.end(); ++it)
        {
            if (job == it->job)
                break;
        }
        if (it == queue.end())
            return;
        queue.erase(it);
    }
}


plumbing::time_value
plumbing::at_service::get_maximum_sleep()
    const
{
    if (queue.empty())
        return time_value(60);
    time_value now = time_value::now();

    //
    // Because we keep the queue sorted, we only ever need to look at
    // the first job in the queue.
    //
    time_value when = queue.front().when;
    if (when <= now)
        return time_value(0);
    return (when - now);
}


void
plumbing::at_service::process()
{
    for (;;)
    {
        if (queue.empty())
            return;

        //
        // Get the current time, and see if first job's time has come.
        //
        time_value now = time_value::now();
        time_value when = queue.front().when;
        if (when > now)
            return;

        //
        // Take the job off the front of the queue *before* we execute
        // it.  This makes the logic simple to re-register itself, or
        // for it to delete itself.
        //
        queue_item item = queue.front();
        queue.pop_front();
        item.job->process();

        //
        // We are not responsible for invoking operator delete on the
        // job object.
        //
    }
}


// vim: ts=4:et
