//
// plumbing - C++ wrapper facades for Berkeley sockets
// Copyright (C) 2009 Peter Miller
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License, version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBPLUMBING_AT_SERVICE_H
#define LIBPLUMBING_AT_SERVICE_H

#include <deque>

#include <libplumbing/time_value.h>

namespace plumbing {

class at_job; // forward

/**
  * The plumbing::at_service class is used to represent a service that runs
  * jobs at a specific time.
  *
  * @note
  *     There is no provision for multi-threading in this code.
  *     The assumption is that all activity is initiated by the
  *     plumbing::reactor::process loop.
  */
class at_service
{
public:
    /**
      * The destructor.
      */
    virtual ~at_service();

    /**
      * The default constructor.
      */
    at_service();

    /**
      * The register_job method is used to add a jobs to the queue of
      * jobs waiting to run.
      *
      * @param when
      *     The time the job is to be called.
      * @param jp
      *     The job to be queued.
      */
    void register_job(const time_value &when, at_job *jp);

    /**
      * The The register_job_delta method is used to add a jobs to the
      * queue of jobs waiting to run.
      *
      * @param seconds
      *     How many seconds into the future the job is to be called.
      * @param jp
      *     The job to be queued.
      */
    void register_job_delta(const time_value &seconds, at_job *jp);

    /**
      * The unregister_job method is used to remove a job from the queue
      * of jobs waiting to run.
      *
      * @param jp
      *     The job to be de-queued.
      */
    void unregister_job(at_job *jp);

    /**
      * The get_maximum_sleep is used to obtain a sleep time, for
      * handing to select(), based on the jobs in the queue.  If there
      * are no jobs, 60 seconds will be returned.
      */
    time_value get_maximum_sleep() const;

    /**
      * The process method will examine all of the jobs in the queue,
      * and run all which have reached or passed their stated execution
      * time.  Jobs are removed from the queue when they are run, but
      * they are NOT deleted.
      *
      * If a job wishes to run again, it is safe to call, and it should
      * call, the at_service::register_job within its at_job::process
      * method.  It is also safe if jobs use "delete this" within their
      * at_job::process method.
      */
    void process();

    /**
      * The empty method is used to determine whether or not the at
      * service still has work to do.
      *
      * @returns
      *     true if no jobs queued, false if there is still work to be done.
      */
    bool empty() const { return queue.empty(); }

private:
    struct queue_item
    {
        time_value when;
        at_job *job;
    };

    typedef std::deque<queue_item> queue_t;

    /**
      * The queue instance variable is used to remember the jobs waiting
      * to run.
      */
    queue_t queue;

    /**
      * The copy constructor.  Do not use.
      */
    at_service(const at_service &);

    /**
      * The assignment operator.  Do not use.
      */
    at_service &operator=(const at_service &);
};

}

#endif // LIBPLUMBING_AT_SERVICE_H
// vim: set ts=4 et
