//
// plumbing - C++ wrapper facades for Berkeley sockets
// Copyright (C) 2009 Peter Miller
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License, version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBPLUMBING_ENDPOINT_H
#define LIBPLUMBING_ENDPOINT_H

#include <boost/shared_ptr.hpp>

#include <cstdio>
#include <string>
#include <libplumbing/format_printf.h>

namespace plumbing {

class time_value; // forward

/**
  * The plumbing::endpoint class is used to represent an object serving
  * a specific client connection, or to represent a clinet's connection
  * to a server; or, more generally, any file descriptor based data
  * source or sink.
  *
  * When endpoints determine that they are ready to be removed from the
  * reactor's list of endpoints and be destroyed, they shall call the
  * #kill_me_now() method.  Depending on when they call this, it may be
  * possible that they will still get some methods called, because the
  * reactor only tests for this condition once per loop.
  */
class endpoint
{
public:
    /**
      * The pointer type is used to present a pointer to an instance of
      * this class.  By defining it here, the pointer semantics can be
      * changed later, in just this one place, and only a recompile is
      * required.
      */
    typedef boost::shared_ptr<endpoint> pointer;

    /**
      * The destructor.
      */
    virtual ~endpoint();

protected:
    /**
      * The constructor.
      * It is protected on purpose, you must derive from this class.
      *
      * @param fd
      *     The file descriptor of the bi-directional socket for
      *     communicating with the client.
      */
    endpoint(int fd);

    /**
      * The constructor.
      * It is protected on purpose, you must derive from this class.
      *
      * @param fd
      *     The file descriptor of the bi-directional socket for
      *     communicating with the client.
      * @param sock_addr
      *     The socket address to use.  Note that it is really of type
      *     sockaddr, but it's void* here so as to avoid unnecessarily
      *     sucking in all of the socket machinery includes.
      * @param sock_addr_len
      *     The length of the socket address data.
      */
    endpoint(int fd, const void *sock_addr, int sock_addr_len);

public:
    /**
      * The get_read_file_descriptor method is used to obtain the file
      * descriptor that select is to wait upon for read events.
      * The default implementation returns the fd instance variable.
      *
      * @returns
      *     int; >=0 for a valid file descriptior, or -1 if read events
      *     should not be waited for.
      *
      * @note
      *     <b>DO NOT</b> close the file descriptor and set it to -1 in
      *     this method.  Do that in the #process_read method instead,
      *     otherwise you confuse the select() processing.
      */
    virtual int get_read_file_descriptor();

    /**
      * The get_write_file_descriptor method is used to obtain the file
      * descriptor that select is to wait upon for write events.
      * The default implementation returns -1.
      *
      * @returns
      *     int; >=0 for a valid file descriptior, or -1 if write events
      *     should not be waited for.
      *
      * @note
      *     This method need not returns the same file descriptor as
      *     returned by the #get_read_file_descriptor method.  You
      *     could, for example, use dup(2) to separate the two.
      *
      * @note
      *     <b>DO NOT</b> close the file descriptor and set it to -1 in
      *     this method.  Do that in the #process_write method instead,
      *     otherwise you confuse the select() processing.
      */
    virtual int get_write_file_descriptor();

    /**
      * The process_read method is called by the select loop when a
      * readable event is detected on the read file descriptor.
      */
    virtual void process_read() = 0;

    /**
      * The process_write method is called by the select loop when a
      * writable event is detected on the read file descriptor.
      */
    virtual void process_write() = 0;

    /**
      * The process timeout method is called by the select loop whenever
      * the select times out without seeing any activity.
      *
      * To choose your timeout, return the desired time from the
      * #get_maximum_sleep method
      *
      * @note
      *     It is possible that a derived class's process_timeout
      *     method will be called, even when it returns "infinity" from
      *     its #get_maximum_sleep method.  ALL endpoints have their
      *     process_timeout method called after select times out.
      */
    virtual void process_timeout();

    /**
      * The is_ready_to_die method is used by the reactor to determine
      * whether or not this endpoint should be removed from the reactor's
      * list of endpoints and destroyed.
      */
    bool is_ready_to_die() const;

    /**
      * The get_maximum_sleep method is used to determine how long
      * this endpoint is prepared to sleep in the absence of any file
      * descriptor activity.
      *
      * Derived classes can call endpoint::get_maximum_sleep() if
      * they need an "infinity" value.  (The default "ininity" is 60
      * seconds.)
      */
    virtual time_value get_maximum_sleep() const;

protected:
    /**
      * The fd instance variable is used to remember the file descriptor
      * of the bi-directional socket for communicating with the client.
      *
      * This instance variable is not private, so that derived classes
      * can manipulate it.  In particular, to cope with network
      * connections going up and down, but providing the illusion of
      * continuous service.
      *
      * Modify this value with extreme care.
      */
    int fd;

    /**
      * The fatal_error method is used to report fatal
      * errors, and then terminate via exit(2).
      *
      * @param fmt
      *     The format string describing the rest of the arguments.  See
      *     printf(3) for more information.
      */
    void fatal_error(const char *fmt, ...)
                                                            FORMAT_PRINTF(2, 3);
    /**
      * The parse_port_number class method is used to translate a
      * textual representation of a port number into an interger.
      *
      * @param text
      *     The port name (e.g. "ssh") or number (e.g. "22").
      * @returns
      *     the port number (1..65525), or -1 on error.
      */
    static int parse_port_number(const std::string &text);

    /**
      * The parse_port_number_or_die class method is used to translate a
      * textual representation of a port number into an interger.
      * If it fails, it exits via the log().fatal_error() method.
      *
      * @param text
      *     The port name (e.g. "ssh") or number (e.g. "22").
      * @param empty_is_ok
      *     This is true if the empty string is acceptable and should
      *     return -1.  This is false of the empty string is not
      *     acceptable and causes a fatal error.
      * @returns
      *     On success, the port number (1..65525); or, if it was the
      *     empty string and empty is OK, -1.  On error, exits via the
      *     log().fatal_error() method.
      */
    static int parse_port_number_or_die(const std::string &text,
        bool empty_is_ok = false);

    /**
      * The fatal_error_v method is used to report fatal errors, and
      * then terminate via exit(2), but it gets its arguments form the
      * given location
      *
      * @param fmt
      *     The format string describing the rest of the arguments.  See
      *     vprintf(3) for more information.
      * @param ap
      *     pointer to the arguments
      */
    void fatal_error_v(const char *fmt, va_list ap)
                                                            FORMAT_PRINTF(2, 0);
    /**
      * The warning method is used to report warnings.
      *
      * @param fmt
      *     The format string describing the rest of the arguments.  See
      *     printf(3) for more information.
      */
    void warning(const char *fmt, ...)
                                                            FORMAT_PRINTF(2, 3);
    /**
      * The warning_v method is used to report warnings, but it gets its
      * arguments form the given location
      *
      * @param fmt
      *     The format string describing the rest of the arguments.  See
      *     vprintf(3) for more information.
      * @param ap
      *     pointer to the arguments
      */
    void warning_v(const char *fmt, va_list ap)
                                                            FORMAT_PRINTF(2, 0);
    /**
      * The set_peer_name method may be used to set the peer_name
      * instance variable from the supplied string.  This method may be
      * called more than once, the lastest value is retained.
      *
      * @param name
      *     The name of the endpoint (usually the other end, in the case
      *     of a network socket).
      */
    void set_peer_name(const char *name);

    /**
      * The set_peer_name method may be used to set the peer_name
      * instance variable from the supplied string.  This method may be
      * called more than once, the lastest value is retained.
      *
      * @param name
      *     The name of the endpoint (usually the other end, in the case
      *     of a network socket).
      */
    void set_peer_name(const std::string &name);

    /**
      * The get_peer_name method may be used to obtain the name of the
      * endpoint (the other end, in the case of network sockets).  This
      * is most often used for error messages.
      */
    const char *get_peer_name() const { return peer_name.c_str(); }

protected:
    /**
      *  The close method is used to close the file descriptor
      *  associated with this end point.  This endpointer will be pretty
      *  much useless after that.
      */
    virtual void close();

    /**
      * The kill_me_now method is used by endpoints when they are
      * ready to be removed from the reactor's list of endpoints and be
      * destroyed.
      */
    void kill_me_now();

private:
    /**
      * The peer_name instance variable is used to remember the name of
      * the other end of the socket.  It is used to give more informative
      * error messages.
      */
    std::string peer_name;

    /**
      * The ready_to_die instance variable is used to remember whether
      * or not this endpoint is ready to be removed from the reactor's
      * list of endpoints and destroyed.
      */
    bool ready_to_die;

    /**
      * The set_peer_name method may be used to probe the socket to
      * discover the identity of the other end of the connection, and
      * set the peer_name instance variable appropriately.
      */
    void set_peer_name();

    /**
      * The set_peer_name method may be used to set the peer_name
      * instance variable from the supplied socket address.
      *
      * @param sock_addr
      *     The socket address to use.  Note that it is really of type
      *     sockaddr, but it's void* here so avoid unnecessarily sucking
      *     in all of the socket machinery includes.
      * @param sock_addr_len
      *     The length of the socket address data.
      */
    void set_peer_name(const void *sock_addr, int sock_addr_len);

    /**
      * The default constructor.  Do not use.
      */
    endpoint();

    /**
      * The copy constructor.  Do not use.
      */
    endpoint(const endpoint &);

    /**
      * The assignment operator.  Do not use.
      */
    endpoint &operator=(const endpoint &);
};

}

#endif // LIBPLUMBING_ENDPOINT_H
