//
// plumbing - C++ wrapper facades for Berkeley sockets
// Copyright (C) 2009 Peter Miller
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License, version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBPLUMBING_ENDPOINT_CLIENT_H
#define LIBPLUMBING_ENDPOINT_CLIENT_H

#include <ctime>
#include <deque>

#include <libplumbing/endpoint.h>

namespace plumbing {

/**
  * The plumbing::endpoint_client class is used to represent a simple text
  * based client, which is able to write to a server and prints all
  * responses from the server on the standard output.
  */
class endpoint_client:
    public endpoint
{
public:
    /**
      * The destructor.
      */
    virtual ~endpoint_client();

protected:
    /**
      * A constructor.
      * It is protected on purpose, you must derive from this class.
      *
      * @param hostname
      *     The name of the remote host.  May be a dotted-quad or a
      *     string to be looked-up via gethostbyname.
      * @param port
      *     The port number.
      */
    endpoint_client(const std::string &hostname, int port);

    /**
      * A constructor.
      * It is protected on purpose, you must derive from this class.
      *
      * @param hostname
      *     The name of the remote host.  May be a dotted-quad or a
      *     string to be looked-up via gethostbyname.
      * @param port
      *     The port number.  May be a string service name ("ssh") or a
      *     numeric port number ("22").
      */
    endpoint_client(const std::string &hostname, const std::string &port);

public:
    /**
      * The write method may be used to write the given text to the
      * server.  A new line will be appended.  The requests will be
      * queued until the connection is established, and then sent when
      * the server is able to accept them.
      *
      * @note
      *    This method does not block.  It returns immediately.  If you
      *    are expecting a response from the server, the #data_received()
      *    method will be called when it arrives.
      *
      *    There is no blocking write method.
      *    There is no blocking read method.
      */
    void write(const std::string &text);

    /**
      * The more_to_send method may be used to determine whether or not
      * the write queue is not empty, that there is still more to send.
      */
    bool more_to_send() const { return !write_queue.empty(); }

    /**
      * The is_connected method may be used to determine whether or not
      * the client is connected to its server or not.
      *
      * @returns
      *     true if connected to server, or false if not connected.
      */
    bool is_connected() const { return (fd >= 0 && !connecting); }

    /**
      * The get_port method may be used to obtain the port number the
      * client is connected to, or will attempt to connect to.
      *
      * @returns
      *     the port number (1 to 65535)
      */
    int get_port() const { return port; }

    /**
      * The get_host method may be used to obtain the name of the host
      * the client is connected to, or will attempt to connect to.
      */
    const std::string &get_host() const { return hostname; }

protected:
    // See base class for documentation.
    int get_read_file_descriptor();

    // See base class for documentation.
    void process_read();

    // See base class for documentation.
    int get_write_file_descriptor();

    // See base class for documentation.
    void process_write();

    /**
      * The data_received method is called by process_read whenever new
      * data arrived from the server.
      *
      * @param data
      *     The data from the server
      * @param nbytes
      *     The size of the data from the server.
      *     This will be zero when the remote server closes the connection.
      */
    virtual void data_received(const void *data, size_t nbytes) = 0;

private:
    /**
      * The hostname instance variable is used to remember which host we
      * are to connect to.
      */
    std::string hostname;

    /**
      * The port instance variable is used to remember which port on the
      * host we are to connect to.
      */
    int port;

    /**
      * The sock_addr instance variable is used to remember the
      * dynamically allocated socket address object.  It is void* rather
      * than drag in all of the <sys/socket.h> baggage.
      */
    void *sock_addr;

    /**
      * The sock_addr_size instance variable is used to remember the
      * size of the dynamically allocated socket address object.
      */
    int sock_addr_size;

    /**
      * The connecting instance variable is used to remember whether we
      * are waiting for a non-blocking connect to complete, or not.
      */
    bool connecting;

    /**
      * The retry_after instance variable is used to remember when to
      * next retry the connect() to the server.
      */
    time_t retry_after;

    /**
      * The write_queue_t typedef describes the type of the write_queue
      * instance variable.  It doesn't include the word "deque" or
      * "vector" so that its implementation may change in the future,
      * and this is the only place which will need editing.
      */
    typedef std::deque<std::string> write_queue_t;

    /**
      * The write_queue instance variable is used to remember the queue
      * of lines waiting to be sent to the server.  This queue may
      * be because the connection has not completed yet, or because
      * previous packets are still being sent.
      */
    write_queue_t write_queue;

    /**
      * The try_to_connect method is used to attempt to the server, if
      * the connection is not currently open.
      */
    void try_to_connect();

    /**
      * The default constructor.  Do not use.
      */
    endpoint_client();

    /**
      * The copy constructor.  Do not use.
      */
    endpoint_client(const endpoint_client &);

    /**
      * The assignment operator.  Do not use.
      */
    endpoint_client &operator=(const endpoint_client &);
};

}

#endif // LIBPLUMBING_ENDPOINT_CLIENT_H
// vim: set ts=8 sw=4 et
