//
// plumbing - C++ wrapper facades for Berkeley sockets
// Copyright (C) 2009 Peter Miller
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License, version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBPLUMBING_ENDPOINT_LISTENER_H
#define LIBPLUMBING_ENDPOINT_LISTENER_H

#include <libplumbing/endpoint.h>

namespace plumbing {

class reactor; // forward

/**
  * The plumbing::endpoint_listener class is used to represent
  * an endpoint which listens for incoming connections.  The
  * endpoint_factory method is then used to allocate a new endpoint to
  * process the new connection.  By creating a listening endpoint, it
  * is possible for a reactor to listen on two or more ports at once,
  * and create appropriate new endpoint derivatived objects for each
  * connection.
  */
class endpoint_listener:
    public endpoint
{
public:
    /**
      * The destructor.
      */
    virtual ~endpoint_listener();

protected:
    /**
      * A constructor.
      * It is protected on purpose, you must derive from this class.
      *
      * @param svc
      *     The reactor to manage all new connections.
      * @param port_number
      *     The port number to use, or -1 if need to ask derived class
      *     for one later.
      * @param repeat_count
      *     The number of time to spawn servers before self terminating.
      *     Use -1 to mean "infinity".
      */
    endpoint_listener(reactor &svc, int port_number, int repeat_count);

    /**
      * A constructor.
      * It is protected on purpose, you must derive from this class.
      *
      * @param svc
      *     The reactor to manage all new connections.
      * @param port_number
      *     The port number to use, or the empty string if need to ask
      *     derived class for one later.
      * @param repeat_count
      *     The number of time to spawn servers before self terminating.
      *     Use -1 to mean "infinity".
      */
    endpoint_listener(reactor &svc, const std::string &port_number,
        int repeat_count);

protected:
    /**
      * The process_read method is called by the select loop when a
      * readable event is detected on the read file descriptor.
      * (This method is virtual in the base class.)
      *
      * Readable events happen when an incoming connection request is
      * detected, and it is handled by the accept() system call to
      * create a new socket to listen to the new file descriptor.
      *
      * A new endpoint is created (via the endpoint-factory method)
      * and added to the managing reactor (via the add_endpoint helper
      * method).
      */
    void process_read();

    // See base class for documentation.
    void process_write();

    /**
      * The add_endpoint method is used to add another endpoint to those
      * being managed by our associated reactor.  This is a simple helper
      * method wrapper, to get the new endpoint to the right place (into
      * the reactor).
      *
      * @param ep
      *     The endpoint to be managed by the reactor.
      */
    void add_endpoint(const endpoint::pointer &ep);

    /**
      * The get_default_port method is used to obtain the port number
      * this reactor will listen on, if not port is specified in the
      * configuration file.  Each drived class must supply its own
      * default port.
      */
    virtual int get_default_port() const = 0;

    /**
      * The endpoint_factory method is call in response to a successful
      * connection by a client.
      *
      * @param fd
      *     The file descriptor of the new connnection
      * @param adr
      *     The address data structure returned by accept
      * @param adrlen
      *     The length of the address data structure returned by accept
      *
      * @returns
      *     A dynamically allocated endpoint instance.
      */
    virtual endpoint::pointer endpoint_factory(int fd, void *adr,
        int adrlen) = 0;

    // See base class for documentation.
    int get_read_file_descriptor();

private:
    /**
      * The svc instance variable is used to remember the reactor to
      * attach new endpoints to, so that they may be handled by the common
      * select() code in the reactor::process() method.
      */
    reactor &svc;

    /**
      * The port_number instance variable is used to remember the port
      * number this endpoint is to listen on.
      *
      * Initially the port number is -1, to indicated that the socket
      * creation has been deferred.  If, when the process method is
      * called, the value is still -1, the get_default_port methd will be
      * used to obtain the port number to be used.
      */
    int port_number;

    /**
      * The repeat_count instance variable is used to remember the
      * number of times to spawn servers before self terminating.
      * Use -1 to mean "infinity".
      */
    int repeat_count;

    /**
      * The create_listening_port method is called by the process
      * method to establish the file descriptor on which to listen for
      * connections.  It is safe to call this method redundantly.
      */
    void create_listening_socket();

    /**
      * The default constructor.  Do not use.
      */
    endpoint_listener();

    /**
      * The copy constructor.  Do not use.
      */
    endpoint_listener(const endpoint_listener &);

    /**
      * The assignment operator.  Do not use.
      */
    endpoint_listener &operator=(const endpoint_listener &);
};

}

#endif // LIBPLUMBING_ENDPOINT_LISTENER_H
// vim: set ts=8 sw=4 et
