//
// plumbing - C++ wrapper facades for Berkeley sockets
// Copyright (C) 2009 Peter Miller
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License, version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

/**
  * @mainpage
  *
  * The Plumbing library provides C++ wrapper facades for Berkeley sockets.
  * It is intended for single-threaded applications.
  *
  * <h2>Major Classes</h2>
  *
  * * The #plumbing::endpoint class hierarchy represents things that
  *   typically have a file descriptor.  This interface defines methods
  *   for reading data when <i>select</i>(2) says it is available for
  *   reading, writing data when <i>select</i>(2) says it is available
  *   for writing, and the necessary ancilliary method to make it
  *   happen.
  *
  * * The #plumbing::endpoint_listener hierarchy are classes that listen
  *   for a connection, and call a factory method (that derived classes
  *   must implement) when a connection is accepted.  These manufactured
  *   endpoints are then also managed by the main loop, all in a single
  *   thread.
  *
  * * The #plumbing::reactor class is used to represent the main loop of
  *   your program.  It monitors endpoints and calls the appropriate
  *   methods for each activity.  When there is no activity, it uses no
  *   CPU, and blocks in the <i>select</i>(2) system call.
  *
  * * It is possible to write network clients using only a
  *   #plumbing::reactor instance, and one or more #plumbing::endpoint
  *   derived classes instances.  Clients can connect to more than one
  *   server simultaneously.
  *
  * * It is possible to write network daemons using
  *   only a plumbing::reactor instance, and one or more
  *   #plumbing::endpoint_listener derived classes, that in turn create
  *   one or more server #plumbing::endpoint derived classes instances.
  *
  * * It is possible to write programs that are both clients <i>and</i>
  *   servers, all managed with a single #plumbing::reactor instance.
  *
  * <h2>Examples</h2>
  *
  * There are several example programs in the sources:
  *
  * * The <i>plumbing-dumb-client</i>(1) command implements a simple
  *   client that reads from a socket and writes everything to stdout.
  *   (See the #plumbing-dumb-client/main.cc file.)
  *
  * * The <i>plumbing-daytime-service</i>(1) command implements a reactor
  *   that listens for daytime (port 13) connections, and replies with
  *   the current time.
  *   (See the #plumbing-daytime-service/main.cc file.)
  *
  * * The <i>plumbing-inet-reactor</i>(1) command implements a TCP-only
  *   equivalent of the traditional <i>inetd</i>(8) super-server.  It is
  *   also able to service daytime (port 13) connections internally, and
  *   some other standard services as well, with only a single thread.
  *   (See the #plumbing-inet-reactor/main.cc file.)
  */

#ifndef LIBPLUMBING_LIBPLUMBING_H
#define LIBPLUMBING_LIBPLUMBING_H

//
// This include file is used by the packaging system, to decide which
// include files are to be installed.  It is also a lazy way for users
// to get all of the plumbing fixtures, at the cost of compile time.
//
#include <libplumbing/at/job.h>
#include <libplumbing/at/service.h>
#include <libplumbing/reactor.h>
#include <libplumbing/endpoint.h>
#include <libplumbing/endpoint/client.h>
#include <libplumbing/endpoint/client/stdout.h>
#include <libplumbing/endpoint/functor.h>
#include <libplumbing/endpoint/listener.h>
#include <libplumbing/logging.h>
#include <libplumbing/logging/file.h>
#include <libplumbing/logging/null.h>
#include <libplumbing/logging/stderr.h>
#include <libplumbing/logging/syslog.h>
#include <libplumbing/logging/tee.h>
#include <libplumbing/time_value.h>

#endif // LIBPLUMBING_LIBPLUMBING_H
