//
// plumbing - C++ wrapper facades for Berkeley sockets
// Copyright (C) 2009 Peter Miller
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License, version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBPLUMBING_LOGGING_H
#define LIBPLUMBING_LOGGING_H

#include <cstdarg>

#include <libplumbing/format_printf.h>

namespace plumbing {

/**
  * The logging class is used to represent an abstract logging
  * destination.  It could be a log file, or it could be syslog, or it
  * could be the standard error stream.
  *
  * Usage: If you do not instantiate a logging instance, a stderr
  * instance will be created at run time.
  *
  * If you want something other than the standard error stream, use the
  * logging::register_stream class method to direct the logging
  * output elsewhere.
  *
  *     logging::register_stream(new logging_file("gwen"));
  *
  * Note that this must be done very early in your program's lifetime.
  */
class logging
{
public:
    /**
      * The destructor.
      */
    virtual ~logging();

protected:
    /**
      * The default constructor.
      *
      * Notice that this constructor is protected.  This is because you
      * must derive from this class in order to create a meaningful
      * logging mechanism.
      */
    logging();

public:
    /**
      * The warning method may be used to log a warning.  The message is
      * delivered via the message_v virtual method.
      *
      * @param fmt
      *     The format of the message, which also describes all
      *     of the following arguments.  See printf(3) for more
      *     information.
      */
    void warning(const char *fmt, ...)                      FORMAT_PRINTF(2, 3);

    /**
      * The warning_v method may be used to log a warning.  The message
      * is delivered via the message_v virtual method.
      *
      * @param fmt
      *     The format of the message.
      * @param ap
      *     the rest of the format's arguments.  See vprintf(3) for more
      *     information.
      */
    void warning_v(const char *fmt, va_list ap)             FORMAT_PRINTF(2, 0);

    /**
      * The fatal_error method may be used to log a fatal error.  The
      * message is delivered via the message_v virtual method.  Program
      * execution is terminated via the exit method.
      *
      * @param fmt
      *     The format of the message, which also describes all
      *     of the following arguments.  See printf(3) for more
      *     information.
      */
    void fatal_error(const char *fmt, ...)         NORETURN FORMAT_PRINTF(2, 3);

    /**
      * The fatal_error_v method may be used to log a fatal error.  The
      * message is delivered via the message_v virtual method.  Program
      * execution is terminated via the exit method.
      *
      * @param fmt
      *     The format of the message.
      * @param ap
      *     the rest of the format's arguments.  See vprintf(3) for more
      *     information.
      */
    void fatal_error_v(const char *fmt, va_list ap)
                                                   NORETURN FORMAT_PRINTF(2, 0);
    /**
      * The get_singleton class method is used to obtain a reference to
      * the sole logging instance for the system.  See the #log() function,
      * below, for how it is used.
      */
    static logging &get_singleton();

    /**
      * The register_stream class method is used to redirect logging
      * into a different logging stream.
      */
    static void register_stream(logging *ls);

    /**
      * The message method is used to format and a print a log message.
      *
      * @param fmt
      *     The format of the message, which also describes all
      *     of the following arguments.  See printf(3) for more
      *     information.
      */
    void message(const char *fmt, ...)                      FORMAT_PRINTF(2, 3);

    /**
      * The message_v method may be used to format and print a log message.
      * The message text is delivered via the deliver_message_text method.
      *
      * @param fmt
      *     The format of the message.
      * @param ap
      *     The rest of the format's arguments.  See vprintf(3) for more
      *     information.
      */
    void message_v(const char *fmt, va_list ap)             FORMAT_PRINTF(2, 0);

protected:
    /**
      * The deliver_message_text method is used to actually deliver a
      * log message to the logging medium.  All formatting has been
      * completed, the text requires no further substitutions.
      */
    virtual void deliver_message_text(const char *txt) = 0;

    /**
      * The exit method is used to terminate the execution of the
      * program.  The exit status will be one (1).
      *
      * @note
      *     This method does not return.
      */
    void exit()                                                        NORETURN;

private:
    /**
      * The singleton class variable is used to remember the address of
      * the sole instance of the logging destination.
      */
    static logging *singleton;

    /**
      * The copy constructor.  Do not use.
      */
    logging(const logging &);

    /**
      * The assignment operator.  Do not use.
      */
    logging &operator=(const logging &);

    // This is so that logging_tee can call deliver_message_text
    // directly, a minor efficiency consideration.
    friend class logging_tee;
};

/**
  * The plumbing::log() function is used to make it possible to simply say
  *
  *     plumbing::log().fatal_error("croak")
  *
  * to issue error messages.  If you need errno translated, use
  * libexplain to do it, and send it through fatal_error().
  * For example
  *
  *     fd = open(filename, flags);
  *     if (fd < 0)
  *         plumbing::log().fatal_error("%s", libexplain_open(filename, flags));
  */
inline logging &log() { return logging::get_singleton(); }

}

#endif // LIBPLUMBING_LOGGING_H
// vim: set ts=8 sw=4 et
