//
// plumbing - C++ wrapper facades for Berkeley sockets
// Copyright (C) 2009 Peter Miller
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License, version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBPLUMBING_REACTOR_H
#define LIBPLUMBING_REACTOR_H

#include <cstdarg>
#include <list>

#include <libplumbing/endpoint.h>
#include <libplumbing/format_printf.h>
#include <libplumbing/at/service.h>

namespace plumbing {

class endpoint_functor; // forward

/**
  * The plumbing::reactor class is used to represent an endpoint manager
  * which dispatches to the appropriate endpoint when some file
  * descriptor activity has been detected for them.
  *
  * The endpoints are completely abstract.  They could be a socket
  * listening for new connections, they could be server processes
  * handling client requests, they could be clients handling
  * asynchronous events, or they could be character devices like serial
  * ports.
  *
  * In particular, if you want your server to listen for connections on
  * a particular port, derive a new class from plumbing::endpoint_listener
  * and define the get_default_port and endpoint_factory methods.  The
  * add this listener to the reactor via the add_endpoint method.  Fram
  * that point, new connection handling is automatic.
  */
class reactor
{
public:
    /**
      * The destructor.
      */
    virtual ~reactor();

    /**
      * The default constructor.
      */
    reactor();

    /**
      * The process method is used to see if any of the if there are
      * any incoming connections, or if any of the current endpoints have
      * input to process.
      *
      * @param block
      *     whether or not we should block
      * @note
      *     This method will only block if a child endpoint blocks.
      */
    void process(bool block = false);

    /**
      * The run method is used to process all file descriptor activity.
      * This method blocks until all connections are closed.
      */
    void
    run()
    {
        while (still_have_work_to_do())
            process(true);
    }

    /**
      * The for_each_endpoint method is used to apply the given functor
      * to each of the connected clients (as presesented by each
      * endpoint object; they could be servers, or clients, or devices,
      * etc).
      *
      * @param func
      *     The operatoin to be performed.
      */
    void for_each_endpoint(endpoint_functor &func);

    /**
      * The add_endpoint method is used to add another endpoint to those
      * being managed by this reactor.  When an endpoint wants to be
      * destroyed, it shall call its kill_Pme_now method, and the reactor
      * will destroy the pointer (and indirectly the object) at the top
      * of its event loop.
      *
      * @param ep
      *     The endpoint to be managed by this reactor.  It shall be
      *     dynamically allocated via operator new, via a class create
      *     method.  It is recommended that this should be the only
      *     pointer to the object once it has been created, otherwise
      *     its destructor may not be called for some time, potentially
      *     leaking resources such as memory and file descriptors.
      */
    void add_endpoint(const endpoint::pointer &ep);

    /**
      * The register_at_job method is used to add a jobs to the queue of
      * jobs waiting to run.
      *
      * @param when
      *     The time the job is to be called.
      * @param jp
      *     The job to be queued.
      */
    void
    register_at_job(const time_value &when, at_job *jp)
    {
        at.register_job(when, jp);
    }

    /**
      * The The register_at_job_delta method is used to add a jobs to
      * the queue of jobs waiting to run.
      *
      * @param when
      *     How many seconds into the future the job is to be called.
      * @param jp
      *     The job to be queued.
      */
    void
    register_at_job_delta(const time_value &when, at_job *jp)
    {
        at.register_job_delta(when, jp);
    }

    /**
      * The unregister_at_job method is used to remove a job from the
      * queue of jobs waiting to run.
      *
      * @param jp
      *     The job to be de-queued.
      */
    void
    unregister_at_job(at_job *jp)
    {
        at.unregister_job(jp);
    }

    /**
      * The still_have_work_to_do method is used to determine whether or
      * not there are any endpoints left to process.
      */
    bool still_have_work_to_do() const { return !endpoints.empty(); }

private:
    typedef std::list<endpoint::pointer> endpoints_t;

    /**
      * The endpoints instance variable is used to remember the
      * currently instantiated endpoints of this reactor.
      */
    endpoints_t endpoints;

    /**
      * The at instance variable is used to remember and manage all of
      * the time-delayed jobs which need to be executed.
      */
    at_service at;

    /**
      * The fatal_error method is used to report fatal errors.
      *
      * @param fmt
      *     The format string describing the rest of the arguments.  See
      *     printf(3) for more information.
      */
    void fatal_error(const char *fmt, ...)
                                                            FORMAT_PRINTF(2, 3);
    /**
      * The fatal_error_v method is used to report fatal errors, but it
      * gets its arguments form the given location
      *
      * @param fmt
      *     The format string describing the rest of the arguments.  See
      *     vprintf(3) for more information.
      * @param ap
      *     pointer to the arguments
      */
    void fatal_error_v(const char *fmt, va_list ap)
                                                            FORMAT_PRINTF(2, 0);

    /**
      * The kill_suicidal_endpoints method is called by the #process
      * method to clean up endpoints that are finished.
      */
    void kill_suicidal_endpoints();

    /**
      * The copy constructor.  Do not use.
      */
    reactor(const reactor &);

    /**
      * The assignment operator.  Do not use.
      */
    reactor &operator=(const reactor &);
};

}

#endif // LIBPLUMBING_REACTOR_H
// vim: set ts=8 sw=4 et
