//
// plumbing - C++ wrapper facades for Berkeley sockets
// Copyright (C) 2009 Peter Miller
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License, version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIBPLUMBING_TIME_VALUE_H
#define LIBPLUMBING_TIME_VALUE_H

#include <sys/time.h>
#include <string>

namespace plumbing {

/**
  * The time_value class is used to represent a time in seconds and
  * microseconds.
  */
class time_value:
    public timeval
{
public:
    /**
      * The destructor.
      *
      * It isn't virtual.  Thou shalt not derive from this class.
      */
    ~time_value() { }

    /**
      * The default constructor.
      *
      * The object will be initialized to exactly 0.0 seconds.
      */
    time_value() { tv_sec = 0; tv_usec = 0; }

    /**
      * The constructor.
      *
      * @param seconds
      *     The time value in seconds
      * @param microseconds
      *     The time value in microseconds.  If this isn't in normal
      *     form (0..999999) it will be normalized - which is why this
      *     method is not inline.
      */
    time_value(long seconds, long microseconds = 0);

    /**
      * The constructor.
      *
      * @param seconds
      *     The time value in seconds
      */
    time_value(int seconds) { tv_sec = seconds; tv_usec = 0; }

    /**
      * The constructor.
      *
      * @param seconds
      *     The value in seconds.  It will be rounded to the closest
      *     microsecond.
      */
    time_value(double seconds);

    /**
      * The constructor.
      *
      * @param str
      *     String to parse for time value.
      *     Will be set to zero on error.
      */
    time_value(const char *str);

    /**
      * The copy constructor.
      */
    time_value(const time_value &);

    /**
      * The assignment operator.
      */
    time_value &
    operator=(const time_value &rhs)
    {
        if (this != &rhs)
        {
            tv_sec = rhs.tv_sec;
            tv_usec = rhs.tv_usec;
        }
        return *this;
    }

    /**
      * The assignment operator.
      */
    time_value &operator=(double);

    /**
      * The assignment operator.
      */
    time_value &
    operator=(long sec)
    {
        tv_sec = sec;
        tv_usec = 0;
        return *this;
    }

    /**
      * The get_seconds method is used to obtain the number of whole
      * seconds in this time value.
      */
    long get_seconds() const { return tv_sec; }

    /**
      * The double operator may be used to cast a time value into a real
      * number of seconds.
      */
    operator double() const;

    /**
      * The get_time_of_day method may be used to set the value of this
      * time to be the value returned by the settimeofdae system call.
      *
      * @returns
      *     The result of the gettimeofday system call:  0 in success,
      *     -1 on failure (with errno set).
      */
    int get_time_of_day();

    /**
      * The now class method may be used to obtain the current time.
      *
      * @note
      *     error reported by get_time_of_day will exit via
      *     plumbing::log.fatal_error - this isn't usually a problem.
      */
    static time_value now();

    /**
      * The representation method is used to form a human-readable
      * representation of the time value.
      */
    std::string representation() const;

    time_value &operator+=(const time_value &);
    time_value operator+(const time_value &) const;
    time_value &operator-=(const time_value &);
    time_value operator-(const time_value &) const;
    time_value operator-() const;
    time_value &operator*=(double);
    time_value operator*(double) const;
    time_value &operator*=(long);
    time_value operator*(long) const;

    bool operator==(const time_value &) const;
    bool operator!=(const time_value &) const;
    bool operator<(const time_value &) const;
    bool operator<=(const time_value &) const;
    bool operator>(const time_value &) const;
    bool operator>=(const time_value &) const;

private:
    /**
      * The in_normal_form method is used within asserts to guarantee
      * that the internal data structure contains value values.
      *
      * @note
      *     The microseconds are always positive.
      *     For example: the value -2.5 is represented as -3 + 0.5
      */
    bool in_normal_form() const { return (tv_usec >= 0 && tv_usec < 1000000); }
};

}

#endif // LIBPLUMBING_TIME_VALUE_H
// vim: set ts=4 et
