//
// plumbing - C++ wrapper facades for Berkeley sockets
// Copyright (C) 2009, 2010 Peter Miller
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License, version 3, as published
// by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <libplumbing/config.h>
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <libexplain/fork.h>

#include <libplumbing/reactor.h>
#include <libplumbing/endpoint/listener/daytime.h>
#include <libplumbing/logging/syslog.h>
#include <libplumbing/progname.h>
#include <libplumbing/version_print.h>


static void
usage()
{
    std::string prog = plumbing::progname_get();
    fprintf(stderr, "Usage: %s [ <option>... ]\n", prog.c_str());
    fprintf(stderr, "       %s -V\n", prog.c_str());
    exit(1);
}


int
main(int argc, char **argv)
{
    std::string port_num;
    plumbing::progname_set(argv[0]);
    bool debug = false;
    int repeat_count = -1;
    for (;;)
    {
        int c = getopt(argc, argv, "adp:V");
        if (c == EOF)
            break;
        switch (c)
        {
        case 'a':
            repeat_count = 1;
            break;

        case 'd':
            debug = true;
            break;

        case 'p':
            port_num = optarg;
            break;

        case 'V':
            plumbing::version_print();
            return 0;

        default:
            usage();
        }
    }
    if (optind != argc)
        usage();

    //
    // The debug command line option says to stay in the foreground, and
    // send errors to stderr.  Otherwise, we redirect logging to syslog.
    //
    if (!debug)
    {
        if (explain_fork_or_die())
            return 0;
        plumbing::logging::register_stream(new plumbing::logging_syslog());
    }

    //
    //
    // This is the simplest reactor main loop that can be achieved with
    // the Plumbing library.
    //
    // The reactor is instantiated, the endpoint is created, and the
    // reactor is told to manage the endpoint.  The endpoint we create
    // is a listener, and when it accepts a connection it creates a new
    // server endpoint.  Now the reactor manages both of them, or as many
    // as are simultaneously connected.  When a server is done, the
    // reactor will destroy it.
    //
    // The reactor then runs until there are no more endpoints to be
    // managed.  Because there will always be a listener, it never returns.
    //
    plumbing::reactor fat_controller;
    plumbing::endpoint_listener::pointer thomas =
        plumbing::endpoint_listener_daytime::create
        (
            fat_controller,
            port_num,
            repeat_count
        );
    fat_controller.add_endpoint(thomas);
    fat_controller.run();
    return 0;
}


// vim: set tw=8 sw=4 et
